#coding: utf-8

from datetime import datetime
import time
import logging

from django.db import transaction

from fias.helpers import iterate_obj_file, fill_object_with_data
from fias.models import (AddrObject, House, NormativeDocument, HouseInterval)
from fias.mappings import (AO_RECORD_MAPPING, HOUSE_RECORD_MAPPING, 
                           NORMDOC_MAPPING, INTERVAL_RECORD_MAPPING)

from threading import Thread, current_thread
from Queue import Queue


logger = logging.getLogger(__name__)


def timer(fn):
    def wrapper(*args, **kwargs):
        logger.debug('"{0}" starting...'.format(fn.__name__))
        t1 = time.time()
        val = fn(*args, **kwargs)
        t2 = time.time()
        logger.debug('"{0}" took {1:,d}s'.format(fn.__name__, int(t2 - t1)))
        return val
    wrapper.__doc__ = fn.__doc__
    wrapper.__name__ = fn.__name__
    return wrapper


@timer
def fill_parents_with_ao(model, batch_size=100, thread_count=4):
    u'''
    Проставление связей в таблице АО на основе поля с GUID родительского объекта.

    Так как для ускорения загрузки мы помещаем объекты в БД пакетами без явного
    проставления ссылок, после этого нам необходимо связать их пост-фактум.

    Алгоритм достаточно простой:
    1. Выбираем все GUID'ы, на которые ссылаются объекты
    2. Получаем для каждого из них ID записи с таким GUID'ом
    3. Проставляем детям каждой записи ссылку на неё одним UPDATE'ом

    NB: Для эффективной работы всего этого нужно все-таки иметь индекс на parent_guid.
    '''
    model_name = model._meta.object_name
    input_q = Queue()
    output_q = Queue()
    records_done = 0

    def updater(in_q, out_q):        
        while True:
            ao_batch = in_q.get()
            if not ao_batch:
                break
            t1 = time.time()
            for ao in ao_batch:
                updated = model.objects.filter(parent_guid=ao.ao_guid).update(parent=ao)
            t2 = time.time()
            out_q.put((current_thread(), len(ao_batch), int(t2 - t1)))
            in_q.task_done()

    logger.info('Filling missing parents for "{0}"'.format(model_name))
    logger.debug('Getting GUIDs with missing parents from "{0}".'.format(model_name))

    parent_guids = model.objects.filter(parent__isnull=True)\
                                     .values_list('parent_guid', flat=True)\
                                     .distinct()
    parents = AddrObject.objects.filter(ao_guid__in=parent_guids,\
                                        actual_status=True)\
                                .only('ao_guid', 'id', 'pk').distinct()

    parents = list(parents)
    total = len(parents)
    total_time = 0
    
    if total > 0:
        threads = [
            Thread(target=updater, 
                   name='Filler-{0}-{1}'.format(model_name, i), 
                   args=(input_q, output_q)) 
            for i in range(thread_count)
        ]

        for t in threads:
            t.start()

        t1 = time.time()
        logger.debug('Filling input queue with {0} records...'.format(total))

        while parents:
            batch, parents = parents[:batch_size], parents[batch_size:]
            input_q.put(batch)

        while True:
            p = output_q.get()
            if p:
                thr, count, t = p
                t2 = time.time()
                records_done += count
                logger.debug('({0}) [{1}s] {2}/{3} [total: {4}s]'.format(thr, t, records_done, total, total_time))
                total_time = t2 - t1

            if records_done == total:
                for i in range(len(threads)):
                    input_q.put([])
                break


@transaction.commit_manually
def load_addrobj_delta(fn):
    u'''
    Загрузка информации об адресных объектах (АО) из файла.
    '''
    # Количество загруженных объектов с разбиением по уровням
    stats = dict.fromkeys([1, 2, 3, 4, 5, 6, 7, 90, 91], 0)

    with open(fn, 'r') as fp:
        for obj in iterate_obj_file(fp, 'Object'):
            try:
                addrobj = AddrObject.objects.get(rec_guid=obj['AOID'])
            except AddrObject.DoesNotExist:
                addrobj = AddrObject()

            fill_object_with_data(addrobj, obj, AO_RECORD_MAPPING)
            addrobj.save()

            stats[addrobj.ao_level] += 1
            del addrobj

    transaction.commit()

    return stats


@transaction.commit_manually
def bulk_load_addrobj(fn, clear=True):
    u'''
    Загрузка информации об адресных объектах (АО) из файла.
    '''
    # Количество загруженных объектов с разбиением по уровням
    stats = dict.fromkeys([1, 2, 3, 4, 5, 6, 7, 90, 91], 0)
    cur_count = 0
    raw_objects = []

    total_time = 0    

    BATCH_SIZE = 50000
    fp = open(fn, 'r')
    file_objects = iterate_obj_file(fp, 'Object')

    if clear:
        # print 'Clearing all objects...'
        AddrObject.objects.all().delete()

    while True:
        prev_ts = time.time()
        obj_dicts = [obj for ind, obj in zip(xrange(BATCH_SIZE), file_objects)]

        for obj in obj_dicts:
            addrobj = AddrObject()
            stats[addrobj.ao_level] = stats.setdefault(addrobj.ao_level, 0) + 1
            fill_object_with_data(addrobj, obj, AO_RECORD_MAPPING)
            raw_objects.append(addrobj)

        AddrObject.objects.bulk_create(raw_objects)            
        cur_count += len(raw_objects)
        raw_objects = []

        t = time.time()
        delta_t = t - prev_ts
        prev = t
        total_time += delta_t
        logger.debug('{0:,d} {1:,d}s (total: {2:,d}s)'.format(cur_count, int(delta_t), int(total_time)))

        if cur_count % BATCH_SIZE != 0:
            break

    # Так как в записях ссылки идут через GUID, нам нужно вручную проставить
    # связи между объектами
    fill_parents_with_ao(AddrObject)

    transaction.commit()
    fp.close()

    return stats


@transaction.commit_manually
def bulk_load_houseobj(fn, clear=True):
    u'''
    Загрузка информации об адресных объектах (АО) из файла.
    '''
    # Количество загруженных объектов с разбиением по уровням
    stats = dict.fromkeys([1, 2, 3, 4, 5, 6, 7, 90, 91], 0)
    cur_count = 0
    raw_objects = []

    total_time = 0    

    BATCH_SIZE = 10000
    fp = open(fn, 'r')
    file_objects = iterate_obj_file(fp, 'House')

    if clear:
        House.objects.all().delete()

    while True:
        prev_ts = time.time()
        obj_dicts = [obj for ind, obj in zip(xrange(BATCH_SIZE), file_objects)]
        read = len(obj_dicts)

        for obj in obj_dicts:
            house = House()
            stats[house.postal_code] = stats.setdefault(house.postal_code, 0) + 1
            fill_object_with_data(house, obj, HOUSE_RECORD_MAPPING)
            raw_objects.append(house)

        House.objects.bulk_create(raw_objects)            
        cur_count += read
        raw_objects = []

        t = time.time()
        delta_t = t - prev_ts
        prev = t
        total_time += delta_t
        logger.debug('{0:,d} {1:,f}s (total: {2:,f}s)'.format(cur_count, delta_t, total_time))

        if read < BATCH_SIZE:
            break

    fill_parents_with_ao(House)
    transaction.commit()
    fp.close()

    return stats


def load_houseobj_delta(fn):
    stats = {}
    i = 0

    with open(fn, 'r') as fp:
        for obj in iterate_obj_file(fp, 'House'):
            try:
                house = House.objects.get(rec_guid=obj['HOUSEID'])
            except House.DoesNotExist:
                house = House()

            fill_object_with_data(house, obj, HOUSE_RECORD_MAPPING)
            try:
                house.parent = AddrObject.actual.get(ao_guid=house.parent_guid)
            except AddrObject.DoesNotExist:
                house.parent = None
            house.save()
            stats[house.postal_code] = stats.setdefault(house.postal_code, 0) + 1

            i += 1
    return stats


def load_normdoc_file(fn):
    stats = {}

    with open(fn, 'r') as fp:
        for obj in iterate_obj_file(fp, 'NormativeDocument'):
            try:
                normdoc = NormativeDocument.objects.get(normdoc_guid=obj['NORMDOCMDOCID'])
            except NormativeDocument.DoesNotExist:
                normdoc = NormativeDocument()

            fill_object_with_data(normdoc, obj, NORMDOC_MAPPING)
            normdoc.save()
            stats[normdoc.doc_type] = stats.setdefault(normdoc.doc_type, 0) + 1
    return stats


def load_interval_file(fn):
    stats = dict.fromkeys([0, 1, 2, 3], 0)

    with open(fn, 'r') as fp:
        for obj in iterate_obj_file(fp, 'HouseInterval'):
            try:
                interval = HouseInterval.objects.get(rec_guid=obj['HOUSEINTID'])
            except HouseInterval.DoesNotExist:
                interval = HouseInterval()

            fill_object_with_data(interval, obj, INTERVAL_RECORD_MAPPING)
            interval.save()
            stats[interval.interval_status] += 1
    return stats


# Соотношение типов файлов с обработчиками, и они
# должны загружаться в определенном порядке
DELTA_LOADERS = {
    'NORMDOC': load_normdoc_file,
    'ADDROBJ': load_addrobj_delta,
    'HOUSE': load_houseobj_delta,
    'HOUSEINT': load_interval_file
}

BULK_LOADERS = {
    'NORMDOC': load_normdoc_file,
    'ADDROBJ': bulk_load_addrobj,
    'HOUSE': bulk_load_houseobj,
    'HOUSEINT': load_interval_file
}
