#coding: utf-8

from django.db.models import Q
from django.http import HttpResponse, HttpResponseNotFound, HttpResponseBadRequest
from django.utils import simplejson
from django.shortcuts import render_to_response
from django.views.decorators.csrf import ensure_csrf_cookie

from fias.models import (AddrObject, House, HouseInterval, REGION, BOROUGH,
                         DISTRICT, CITY, INTRACITY, PLACE, STREET,
                         ADD_TERRITORY, SUB_ADD_TERRITORY)


class JsonResponse(HttpResponse):
    def __init__(self, content, mimetype='application/json',
                 status=None, content_type=None):
        super(JsonResponse, self).__init__(
            content=simplejson.dumps(content),
            mimetype=mimetype,
            status=status,
            content_type=content_type,
        )


def get_ao_object(request, object_id):
    result = {}

    # Запрашиваем по GUID'у адресного объекта
    query_field = None
    id_len = len(object_id)
    if id_len == 36:
        query_field = 'ao_guid'
    elif id_len == 13:
        query_field = 'kladr_code'

    if not query_field:
        return HttpResponseBadRequest()

    try:
        obj = AddrObject.actual.get(**{query_field: object_id})
    except AddrObject.DoesNotExist:
        return HttpResponseNotFound()

    result = {
        '_record': obj.rec_guid,
        'guid': obj.ao_guid,
        'parent_guid': obj.parent_guid,
        'postcode': obj.postal_code,
        'ifnsfl': obj.ifnsfl,
        'ifnsul': obj.ifnsul,
        'terr_ifnsfl': obj.terr_ifnsfl,
        'terr_ifnsul': obj.terr_ifnsul,
        'okato': obj.okato,
        'oktmo': obj.oktmo,
        'level': obj.ao_level,
        'short_name': obj.shortname,
        'formal_name': obj.formal_name,
        'official_name': obj.official_name,
        'updated': obj.update_date.strftime('%d.%m.%Y'),
        'address': obj.address_text()
    }

    return JsonResponse(result)


def get_house_object(request, guid):
    result = {}

    try:
        obj = House.objects.get(house_guid=guid)
    except AddrObject.DoesNotExist:
        return HttpResponseNotFound()

    result = {
        'ao_guid': obj.house_guid,
        'parent_guid': obj.parent.ao_guid,
        'postcode': obj.postal_code,
        'ifnsfl': obj.ifnsfl,
        'ifnsul': obj.ifnsul,
        'terr_ifnsfl': obj.terr_ifnsfl,
        'terr_ifnsul': obj.terr_ifnsul,
        'okato': obj.okato,
        'oktmo': obj.oktmo,
        'updated': obj.update_date.strftime('%d.%m.%Y'),
        'address': obj.address_text()
    }

    return JsonResponse(result)


def search_ao_objects(request):
    result_fields = ['rec_guid', 'ao_level', 'ao_guid', 'shortname',
                     'formal_name', 'official_name', 'postal_code',
                     'parent__formal_name', 'parent__shortname']
    query_fields = ['official_name', 'shortname']
    search_query = Q()
    # Уровни выводимых АО
    levels = range(1, 8)

    # Проверяем на наличие обязательного параметра
    text_query = request.GET.get('filter', request.POST.get('filter'))
    if not text_query:
        return HttpResponseBadRequest("Missing parameter: 'filter'")

    # Проверяем, не нужно ли ограничить выборку границами какого-либо АО
    boundary_guid = request.POST.get('boundary', None)
    if boundary_guid:
        code_fields = [
            'region_code',
            'autonomy_code',
            'area_code',
            'city_code',
            'ctar_code',
            'place_code',
            'street_code'
        ]

        try:
            bound_ao = AddrObject.actual.get(ao_guid=boundary_guid)
        except AddrObject.DoesNotExist:
            return HttpResponseBadRequest('Requested boundary object "{0}" does not exist!'.format(boundary_guid))

        bound_query = Q()
        for field in code_fields[:bound_ao.ao_level]:
            bound_query &= Q(**{field: getattr(bound_ao, field)})
        search_query &= bound_query


    # Если нужно, то ограничим выборку только объектами выбранных уровней
    try:
        levels = [int(l) for l in request.POST.getlist('levels', levels)]
        search_query &= Q(ao_level__in=levels)
    except ValueError, e:
        return HttpResponseBadRequest("Invalid AO level: {0}".format(e.message))

    words_filter = Q()
    words = text_query.split(' ')
    # Убираем пустые элементы, образованные из лишних пробелов
    words = map(lambda x: x, words)

    for word in words:
        field_filter = Q()
        # Ищем совпадения с указанными словами (частями слов) в
        # нескольких полях сразу (sic!)
        for field in query_fields:
            field_filter |= Q(**{field + '__icontains': word})
        words_filter &= field_filter

    found_objects = AddrObject.actual.filter(search_query & words_filter)\
                                     .select_related('parent__formal_name', 'parent__short_name')\
                                     .order_by('ao_level', 'formal_name')\
                                     .values(*result_fields)[0:50]

    # Если не найдены объекты, содержащие все слова из фильтра,
    # то выберем все объекты, у которых есть родители с хотя бы
    # одним из указанных слов в официальном наименовании. (sic!)
    if not found_objects:
        words_filter = Q()
        for word in words:
            compound_filter = Q()
            for field in query_fields:
                field_filter = Q(**{field + '__icontains': word}) |\
                               Q(**{'parent__' + field + '__icontains': word}) |\
                               Q(**{'parent__parent__' + field + '__icontains': word}) |\
                               Q(**{'parent__parent__parent__' + field + '__icontains': word})
                compound_filter |= field_filter
            words_filter &= compound_filter

        found_objects = AddrObject.actual.filter(search_query & words_filter)\
                                         .select_related('parent__formal_name', 'parent__short_name')\
                                         .values(*result_fields)[0:50]
    results = list(found_objects)
    total = len(results)

    for r in results:
        parent_part = ''
        if r['parent__formal_name']:
            parent_part = u' / {0}. {1}'.format(r['parent__shortname'], r['parent__formal_name'])
        r['name'] = u'{0}. {1}{2}'.format(r['shortname'], r['formal_name'], parent_part)

    results = sorted(results, key=lambda r: (r['ao_level'], r['shortname'], r['formal_name']))

    return JsonResponse(dict(total=total, rows=results))

def list_house_suggestions(request):
    part = request.POST.get('part')
    street_guid = request.POST.get('street')
    houses = []

    houses = House.objects.filter(parent_guid=street_guid)\
                          .filter(house_number__startswith=part)\
                          .order_by('house_number')\
                          .values('house_number', 'postal_code', 'house_guid')

    return JsonResponse(dict(total=len(houses), rows=list(houses)))


def search_form(request):
    return render_to_response('search_example.html')
